﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Algorithms.DataAnalysis.SymRegGrammarEnumeration {
  [StorableClass]
  public class SymbolList : DeepCloneable, IList<Symbol> {
    [Storable]
    private List<Symbol> symbols;

    public int Count { get { return symbols.Count; } }

    public int TerminalSymbolCount { get { return symbols.Count(s => s is TerminalSymbol); } }
    public int NonTerminalSymbolCount { get { return symbols.Count(s => s is NonterminalSymbol); } }

    public bool IsReadOnly {
      get { return false; }
    }

    /// <summary>
    /// Return the complexity as the number of non-terminal or variable-terminal symbols
    /// </summary>
    public int Complexity {
      get {
        int c = 0;
        foreach (var s in symbols) {
          if (s is NonterminalSymbol || s is VariableTerminalSymbol) c++;
        }
        return c;
      }
    }

    public Symbol this[int index] {
      get { return symbols[index]; }
      set { symbols[index] = value; }
    }

    public SymbolList(params Symbol[] symbols) {
      this.symbols = symbols.ToList();
    }

    public SymbolList(IEnumerable<Symbol> symbols) {
      this.symbols = symbols.ToList();
    }

    public SymbolList(List<Symbol> symbols) {
      this.symbols = symbols;
    }

    [StorableConstructor]
    protected SymbolList(bool deserializing) { }

    protected SymbolList(SymbolList original, Cloner cloner) {
      symbols = original.symbols.Select(cloner.Clone).ToList();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolList(this, cloner);
    }

    public override string ToString() {
      return string.Join(" ", this);
    }

    public bool IsSentence() {
      return !this.Any(sym => sym is NonterminalSymbol);
    }

    public int[] GetNonterminalSymbolIndexes() {
      return Enumerable.Range(0, symbols.Count)
        .Where(i => symbols[i] is NonterminalSymbol)
        .ToArray();
    }

    public int NextNonterminalIndex() {
      Debug.Assert(!IsSentence());
      int firstNonTerminalIndex = 0;
      while (firstNonTerminalIndex < symbols.Count && symbols[firstNonTerminalIndex] is TerminalSymbol) {
        firstNonTerminalIndex++;
      }
      return firstNonTerminalIndex;
    }

    public SymbolList DerivePhrase(int nonTerminalSymbolIndex, SymbolList production) {
      // derive a new phrase by replacing the non-terminal at the given index with the production
      var derived = new List<Symbol>(Count + production.Count - 1);
      derived.AddRange(symbols.Take(nonTerminalSymbolIndex));
      derived.AddRange(production);
      derived.AddRange(symbols.Skip(nonTerminalSymbolIndex + 1));
      return new SymbolList(derived);
    }

    #region IList<Symbol> methods
    public IEnumerator<Symbol> GetEnumerator() {
      return symbols.GetEnumerator();
    }

    IEnumerator IEnumerable.GetEnumerator() {
      return GetEnumerator();
    }

    public int IndexOf(Symbol item) {
      return symbols.IndexOf(item);
    }

    public void Insert(int index, Symbol item) {
      symbols.Insert(index, item);
    }

    public void RemoveAt(int index) {
      symbols.RemoveAt(index);
    }

    public void Add(Symbol item) {
      symbols.Add(item);
    }

    public void Clear() {
      symbols.Clear();
    }

    public bool Contains(Symbol item) {
      return symbols.Contains(item);
    }

    public void CopyTo(Symbol[] array, int arrayIndex) {
      symbols.CopyTo(array, arrayIndex);
    }

    public void CopyTo(int index, Symbol[] array, int arrayIndex, int count) {
      symbols.CopyTo(index, array, arrayIndex, count);
    }

    public bool Remove(Symbol item) {
      return symbols.Remove(item);
    }
    #endregion
  }
}
